        PAGE 55, 132
        TITLE  PC SASI DEVICE DRIVER  REV 1.1
;-----------------------------------------------------------------*
;  MSDOS DEVICE DRIVER: SASI FIXED DISK INTERFACE.
;
;       Copyright 1985,86 by Turning Point Logic Corporation.
;       Permission granted for personal, non-commercial use.
;
;       The Shugart Associates System Interface, better known as
;       SASI, is a very popular hardware/software standard for
;       hard disks.  The interface, both on hardware and software
;       terms, is fairly easy to implement and supports devices
;       such as hard disks, floppy disks, printers, optical storage
;       disks, and other devices with little or no software changes.
;
;       The SASI interface has become an ANSI standard called Small
;       Computer Systems Interface, referred to as SCSI (usually
;       pronouced "SCUZZY" - although the pronoucation does not
;       reflect its capabilities).
;
;       In the area of hard disks, for which SASI was originally
;       designed, disks from 5 to 512 megabytes are supported.  
;       The SASI interface allows for up to eight hard disk units
;       referred to as "controllers".  Each controller can support 
;       at least one hard disk.  Therefore, for those in need of 
;       massive disk space a user can configure a system with 4 
;       gigabytes of disk space (512MB/disk * 8 controllers).
;
;       This driver is installed at load time by PC-DOS by including
;       the statement: DEVICE = SASI.EXE  in the CONFIG.SYS file.
;
; ----- REQUEST HEADER STRUCTURE.
;
RHST    STRUC
;
RHLTH   DB ?;                           Request header length.
RHUCD   DB ?;                           Unit code.
RHCMD   DB ?;                           Command code.
RHSTAT  DW ?;                           Status word.
RHRSV   DB 8 DUP(?);                    Reserved DOS area.
RHDATA  DB 20 DUP(?);
;
RHST    ENDS
;
; ----- BIOS PARAMETER BLOCK STRUCTURE.
;
BPB     STRUC
;
BPBBS   DW ?;                           Bytes per sector.
BPDSU   DB ?;                           Sector per allocation unit.
BPBLS   DW ?;                           Last reserved logical
;                                        sector.
BPBNF   DB ?;                           Number of FATs.
BPBRD   DW ?;                           Number of root directories.
BPBIS   DW ?;                           Number of sectors in logical
;                                        image.
BPBMD   DB ?;                           Media descriptor byte.
BPBFS   DW ?;                           Number of sector per FAT.
BPBST   DW ?;                           Number of sectors per track.
BPBNH   DW ?;                           Number of heads.
BPBHS   DW ?;                           Number of hidden sectors.
;
BPB     ENDS
;
; ----- ADDITIONAL DRIVE CHARACTERISTICS PARAMETERS STRUCTURE.
;
ADCST   STRUC
;
ADCYM   DW ?;                           Number of cylinders.
ADILV   DW ?;                           Interleave factor.
ADCRC   DW ?;                           Starting cylinder number 
;                                        for reduced current.
ADCWP   DW ?;                           Starting cylinder number
;                                        for write precomp.
ADECC   DW ?;                           Maximum ECC data burst
;                                        length.
ADSCV   DW ?;                           SASI DCB Control byte value.
;
ADCST   ENDS
;
; ----- INITIALIZATION REQUEST BLOCK STRUCTURE.
;
RINIT   STRUC
;
IRH     DB 13 DUP(?);                   Request header.
IUNUM   DB ?;                           Number of units supported.
ICENDL  DW ?;                           Ptr to end of code [offset]
ICENDH  DW ?;                                              [segment]
IBPBL   DW ?;                           Ptr to BPB ptr tbl [offset]
IBPBH   DW ?;                                              [segment]
;
RINIT   ENDS
;
; ----- MEDIA CHECK REQUEST BLOCK STRUCTURE.
;
MDCK    STRUC
;
MRH     DB 13 DUP(?);                   Request header.
MDB     DB ?;                           Media descriptor byte.
MRB     DB ?;                           Return code.
;
MDCK    ENDS
;
; ----- BUILD BPB REQUEST BLOCK STRUCTURE.
;
BRBPB   STRUC
;
BRRH    DB 13 DUP(?);                   Request header.
BRMDB   DB ?;                           Media descriptor byte.
BRDTL   DW ?;                           Data transfer address (low)
BRDTH   DW ?;                           Data transfer address (high)
BRPTRL  DW ?;                           Pointer to BPB table (low)
BRPTRH  DW ?;                           Pointer to BPB table (high)
;
BRBPB   ENDS
;
; ----- INPUT/OUTPUT REQUEST BLOCK STRUCTURE.
;
IOSTR   STRUC
;
IORH    DB 13 DUP(?);                   Request header.             
IOMDB   DB ?;                           Media descriptor byte.      
IODTAL  DW ?;                           Data transfer address (low) 
IODTAH  DW ?;                           Data transfer address (high)
IOCNT   DW ?;                           Number of sectors actually
;                                        transferred.
IOSECT  DW ?;                           Starting sector number.
;
IOSTR   ENDS
;
; ----- HARD DISK DRIVER PARAMETER STRUCTURE.
;
HDPBS   STRUC
;
HCTYPE  DW ?;                           Call type code.
HCMMD   DW ?;                           Command code.
HCNUM   DW ?;                           Drive number (0-7).
HCCNT   DW ?;                           Sector count (total of 32K
;                                        bytes maximum).
HCSCTL  DW ?;                           Starting logical sector
HCSCTH  DW ?;                            number [low word, high word]
HDTAL   DW ?;                           Data transfer address
HDTAH   DW ?;                            [low word, high word].
HINLV   DW ?;                           Interleave factor (init).
HCYNM   DW ?;                           Max # cylinders (init).
HHNM    DW ?;                           Number of heads (init).
HRWCC   DW ?;                           Starting cylinder for
;                                        write current (init).
HWPC    DW ?;                           Starting cylinder for
;                                        write precomp (init).
HECCM   DW ?;                           Maximum ECC data burst
;                                        length (init).
HDCBC   DW ?;                           SASI DCB Control byte value
;                                        for options (init).
HSCTK   DW ?;                           Number of sectors per track.
;
HDPBS   ENDS
;
;
PROG    SEGMENT BYTE    PUBLIC  'PROG'
        ASSUME CS:PROG
SASI    PROC FAR
;
; ----- EXTERNALS.
;
EXTRN   TDBUF:NEAR
EXTRN   IRSASI:NEAR
EXTRN   HDDRV:NEAR
;
; ----- EQUATES.
;
HD_SLTH EQU 34;                         Length of the disk driver
;                                        parameter block.
BPB_LTH EQU 19;                         Length of the BIOS para-
;                                        meter block.
ADC_LTH EQU 12;                         Length of the Additional
;                                        Drive Characteristic data.
RH_CMAX EQU 12;                         Highest device driver
;                                        function code.
RT_CNT  EQU 3;                          Retry count.
COR_ERR EQU 18H;                        Correctable Data Error code
;                                        returned by low level SASI
;                                        disk driver.
;
BPB_OFFS EQU 11;                        Offset into the boot sector
;                                        to the BPB data.
ADC_OFFS EQU 44;                        Offset into the boot sector
;                                        to the additional drive
;                                        characteristics data.
;
; ----- DEVICE HEADER BLOCK.
;
DHDR    DD -1;                          Forward device driver
;                                        link pointer.  Set by
;                                        DOS at load time.
        DW 0000000000000000B;           Attribute word
        DW STRAT;                       Pointer to device strategy
;                                        entry point.
        DW IENTRY;                      Pointer to device inter-
;                                        rupt entry point.
DUCNT   DB 1;                           Number of units supported
;                                        (block device).
        DB 7 DUP(0);                     Remainder of name/unit
;                                        field.
;
; ----- BIOS PARAMETER BLOCK POINTER TABLE.
;
BPBPT   DW BPB0;                        Pointer to subunit Bios
        DW BPB1;                         Parameter Blocks.
        DW BPB2;
        DW BPB3;                
;
BPB0    DW BPB_LTH DUP(?);
BPB1    DW BPB_LTH DUP(?);
BPB2    DW BPB_LTH DUP(?);
BPB3    DW BPB_LTH DUP(?);
;
; ----- ADDITIONAL DRIVE CHARACTERISTIC POINTER TABLE.
;
ADCTB   DW ADC0;
        DW ADC1;
        DW ADC2;
        DW ADC3;
;
ADC0    DB ADC_LTH DUP(?);
ADC1    DB ADC_LTH DUP(?);
ADC2    DB ADC_LTH DUP(?);
ADC3    DB ADC_LTH DUP(?);
;
; ----- STACK AREA.
;
        DW 127 DUP(?);
STAREA  DW ?;
;
; ----- REQUEST HEADER POINTER [Set by stradegy routine].
;
RHPTR   LABEL DWORD;
RHOFF   DW ?;                           Request header segment off-
;                                        set.
RHSEG   DW ?;                           Request header segment base.
;
; ----- GENERAL VARIABLES.
;
STPTR   DW ?;                           Stack pointer value on entry.
STSEG   DW ?;                           Stack segment value on entry.
; 
CNUM    DW ?;                           Loop control for initializa-
;                                        tion routine.
IOCMMD  DW ?;                           Read/write control byte,
;                                        (used for looping for
;                                         verify operation).                                    
;
; ----- FUNCTION JUMP TABLE.
;
FUNTAB  DW SINIT;                       Hardware and software ini-
;                                        tialization.
        DW MCHK;                        Media check.
        DW BLDBPB;                      Build Bios Parameter Block.
        DW NVCMD;                       IOCTL input.
        DW SREAD;                       Read block.
        DW NVCMD;                       Non-destructive Input status
;                                        (char only).
        DW NVCMD;                       Input Status (char only).
        DW NVCMD;                       Input flush (char only).
        DW SWRITE;                      Write block.
        DW SVWRT;                       Write block w/verification.
        DW NVCMD;                       Output status (char only).
        DW NVCMD;                       Output flush (char only).
        DW NVCMD;                       IOCTL output.
;
; ----- STRATEGY ENTRY POINT.
;
STRAT:  MOV CS:RHSEG, ES;               Save the segment and offset
        MOV CS:RHOFF, BX;                of the request header and
        RET;                             return.
;
; ----- INTERRUPT ENTRY POINT.
;
IENTRY: PUSH AX;                        Save work registers.
        PUSH BX;
        PUSH CX;
        PUSH DX;
        PUSH DI;
        PUSH SI;
        PUSH DS;
        PUSH ES;
        PUSH BP;
;
        PUSHF;                          Save the flags which
        POP CX;                          includes the current
        PUSH CX;                         interrupt enable
;                                        setting.
;
        MOV CS:STSEG, SS;               Save the current stack 
        MOV CS:STPTR, SP;                segment and pointer.
;

        CLI;                            Set the stack pointer to
        MOV AX, CS;                      a new area and reset
        MOV BX, OFFSET STAREA;           the interrupt state.
        MOV SS, AX;
        MOV SP, BX;
        PUSH CX;
        POPF;
;
        LDS BX, CS:RHPTR;               Load pointer to request
;                                        header.
;
        MOV AL, [BX].RHCMD;             Load command code and
        CMP AL, RH_CMAX;                 check its validity.
        JBE ISIND;
        CALL NVCMD;
        JMP ISRTN;
;
ISIND:  XOR AH, AH;                     Use the command code as an 
        SHL AX, 1;                       index into a jump table.
        MOV DI, AX;
;
        CALL WORD PTR CS:FUNTAB[DI];    Proceed to process request.
;
ISRTN:  MOV SS, CS:STSEG;               Restore the original    
        MOV SP, CS:STPTR;                stack pointer.         
;
        POPF;                           Restore flags which
;                                        sets the interrupt enable
;                                        to what it was on entry.
;
        POP BP;                         Restore work registers.
        POP ES;
        POP DS;
        POP SI;
        POP DI;
        POP DX;
        POP CX;
        POP BX;
        POP AX;
;
        RET;                            Return.
;
;
; ----- DRIVER INITIALIZATION.
;
;       ENTRY: DS:BX points to the request header.
;              RHOFF and RHSEG contain the request header address.
;
;       EXIT:  The specified operation has been performed.
;              The status word in the RH has been set to the
;               appropriate value.
;              All registers have been destroyed.
;
;
SINIT   PROC NEAR
;
        CALL IRSASI;                    Reset all of the controllers.
;
        SUB SP, HD_SLTH;                Create room on the stack
        MOV BP, SP;                      for driver parameters.
;
        MOV AL, CS:DUCNT;               Set the number of units
        MOV [BX].IUNUM, AL;              from device header value.
;
        XOR AH, AH;                     Set loop counter.
        MOV CS:CNUM, AX;
;
IDLP:   MOV [BP].HCTYPE, 1;             Set caller type code.
;
        MOV [BP].HCMMD, 0;              Set read sector command code.
;
        MOV [BP].HCCNT, 1;              Set sector count to one.
;
        MOV AX, CS:CNUM;                Set drive number.
        MOV AL, AH;
        XOR AH, AH;
        MOV [BP].HCNUM, AX;             
;
        MOV [BP].HCSCTL, 0;             Set logical sector zero for
        MOV [BP].HCSCTH, 0;              read operation.
;
;
        MOV [BP].HDTAL, OFFSET TDBUF;   Set data buffer pointer to
        MOV [BP].HDTAH, CS;              temporary buffer at end of
;                                        device driver.
;
        MOV CX, RT_CNT;                 Set the retry count and
IRBT:   CALL HDDRV;                      perform the read operation.
        CMP AX, 0;                       If an error occurred, reset
        JE ICBT;                         the SASI bus, and try again
        CMP AX, COR_ERR;                 until the retry count has   
        JE ICBT;                         been exhausted.             
        LOOP IRBT;
        JMP INXTD;
;
ICBT:   MOV CX, BPB_LTH;                Copy the Bios Parameter Block
        MOV SI, OFFSET TDBUF+BPB_OFFS;   values from the boot loader
        MOV AX, CS:CNUM;                 sector into the BPB.
        MOV AL, AH;
        XOR AH, AH;
        SHL AX, 1;
        MOV DX, AX;
        MOV DI, OFFSET BPBPT;
        ADD DI, AX;
        MOV DI, CS:[DI];
        MOV AX, CS;
        MOV ES, AX;
        MOV DS, AX;
        REP MOVSB;
;
        MOV CX, ADC_LTH;                Copy the additional drive
        MOV SI, OFFSET TDBUF+ADC_OFFS;   characteristic information
        MOV DI, OFFSET ADCTB;            values from the boot loader
        ADD DI, DX;                      sector to the data area.
        MOV DI, CS:[DI];
        REP MOVSB;
;
        MOV DS, CS:RHSEG;               Restore segment portion of
;                                        request header pointer.
;
        CALL ISDRVCHAR;                 Issue the drive charact-
;                                        erisics to the SASI
;                                        controller.
;
INXTD:  MOV AX, CS:CNUM;                Loop for all units.
        INC AH;
        DEC AL;
        MOV CS:CNUM, AX;
        JZ ISBPB;
        JMP IDLP;
;
ISBPB:  MOV [BX].IBPBL, OFFSET BPBPT;   Set the pointer to the BPB
        MOV [BX].IBPBH, CS;              pointer table in the RH.
;
        MOV [BX].ICENDL, OFFSET TDBUF;  Set pointer to the end of
        MOV [BX].ICENDH, CS;             the device driver.
;
        MOV AX, 0100H;                  Set success status.
;
        MOV [BX].RHSTAT, AX;            Store the status code into
;                                        the request header.
;
        ADD SP, HD_SLTH;                Remove the disk driver
;                                        parameters from the stack.
;
        RET;                            Return.
;
SINIT   ENDP
;
; ----- ISSUE DRIVE CHARACTERISTICS SUBROUTINE.
;
;       ENTRY: The BPB values and the additional drive
;               characteristics must be valid.
;              The most signficant byte of CNUM contains
;               the drive number.
;
;       EXIT:  AX contains a result as returned by the disk driver.
;
ISDRVCHAR  PROC NEAR
;
        PUSH BP;                        Save work registers.
        PUSH BX;
        PUSH CX;
        PUSH DX;
;
        MOV AX, CS:CNUM;                Load the local copy of
        MOV AL, AH;                      the controller number.
        XOR AH, AH;                      From the controller number
        MOV DX, AX;                      calculate an index into the
        SHL AX, 1;                       Additional-Drive-Char-
        MOV CX, AX;                      acteristics Table and the
;                                        the Bios Parameter Block 
;                                        for the specified con-
;                                        troller.
;
        SUB SP, HD_SLTH;                Create room on the stack
        MOV BP, SP;                      for driver parameters.
;
        MOV [BP].HCTYPE, 1;             Set caller type,
        MOV [BP].HCMMD, 2;               command code,
        MOV [BP].HCNUM, DX;              SASI controller number,
        MOV BX, OFFSET BPBPT;            [set pointer to the BPB
        ADD BX, CX;                       for the specified con-
        MOV BX, CS:[BX];                  troller].
        MOV AX, CS:[BX].BPBNH;           number of heads,
        MOV [BP].HHNM, AX;
        MOV AX, CS:[BX].BPBST;           number of sectors per track,
        MOV [BP].HSCTK, AX;
        MOV BX, OFFSET ADCTB;            [set pointer to the
        ADD BX, CX;                       Additional Drive Character-
        MOV BX, CS:[BX];                  istics for specified cntrl].
        MOV AX, CS:[BX].ADILV;           interleave factor,
        MOV [BP].HINLV, AX;
        MOV AX, CS:[BX].ADCYM;
        MOV [BP].HCYNM, AX;              number of cylinders,
        MOV AX, CS:[BX].ADCRC;           cylinder number for reduced
        MOV [BP].HRWCC, AX;               write current,
        MOV AX, CS:[BX].ADCWP;           cylinder number for write
        MOV [BP].HWPC, AX;                precompensation,
        MOV AX, CS:[BX].ADECC;           ECC data burst length.
        MOV [BP].HECCM, AX;
        MOV AX, CS:[BX].ADSCV;           SASI DCD Control byte value.
        MOV [BP].HDCBC, AX;
;                               
        MOV CX, RT_CNT;                 Set the retry count and
ISDC:   CALL HDDRV;                      perform the operation.
        CMP AX, 0;                       If an error occurred, try
        JE ISCRTN;                       again until the retry count
        LOOP ISDC;                       has been exhausted.
;
ISCRTN: ADD SP, HD_SLTH;                Remove the disk drive para-
;                                        meters from the stack.
;
        POP DX;                         Restore work registers.
        POP CX;
        POP BX;
        POP BP;
;
        RET;                            Return.
;
ISDRVCHAR  ENDP
;
; ----- PERFORM MEDIA CHECK REQUEST.
;
;       ENTRY: DS:BX points to the request header.
;
;       EXIT:  The specified operation has been performed.
;              The status word in the RH has been set to the
;               appropriate value.
;              The registers, AX, CX, DX have been destroyed.
;
;
MCHK    PROC NEAR
;
;
        PUSH BP;                        Save work registers.
;
        SUB SP, HD_SLTH;                Create room on the stack
        MOV BP, SP;                      for driver parameters.
;
        MOV [BP].HCTYPE, 1;             Set caller type,
        MOV [BP].HCMMD, 5;               command code,
        MOV AH, [BX].RHUCD;              unit number.
        XOR AL, AL;
        MOV CS:CNUM, AX;
        XCHG AL, AH;
        MOV [BP].HCNUM, AX;
;
        MOV CX, RT_CNT;                 Set retry count.
;
MCLP:   CALL HDDRV;                     Perform a test drive ready
;                                        operation.
;
        CMP AX, 0;                      If the test drive ready
        JNE MCDC;                        operation indicated that
        MOV [BX].MRB, 1;                 the media has not 
        MOV AX, 0100H;                   changed then set that
        JMP MCRTN;                       status into the RH and
;                                        exit with success code.
;
MCDC:   CMP AX, 1;                      If the test drive ready
        JNE MCERR;                       operation indicated that
        MOV [BX].MRB, -1;                the media has changed (or
        MOV AX, 0100H;                   been corrupted by another
        JMP MCRTN;                       host) then set that status
;                                        into the RH and exit with
;                                        success code.
;
MCERR:  DEC CX;                         Otherwise, an error has 
        JZ MCERR1;                       occurred.  Retry until the
        CALL RSCTL;                      retry count has been ex-
        JMP MCLP;                        hausted, resetting the SASI
;                                        controller in between.
;
MCERR1: MOV [BX].MRB, 0;                If the error is not recover-
        MOV AX, 800CH;                   able, then set in the RH 
;                                        that it is not known if the
;                                        media changed and set and
;                                        error result code.
;
MCRTN:  MOV [BX].RHSTAT, AX;            Set status code into the
;                                        Request Header.
;
        ADD SP, HD_SLTH;                Remove the disk drive para-
;                                        meters from the stack.
;
        POP BP;                         Restore work register.
;
        RET;                            Return.
;                                        
MCHK    ENDP
;
; ----- PERFORM BUILD BPB REQUEST.
;
;       ENTRY: DS:BX points to the request header.
;
;       EXIT:  The specified operation has been performed.
;              The status word in the RH has been set to the
;               appropriate value.
;              The registers, AX, CX, DX have been destroyed.
;
;
BLDBPB  PROC NEAR
;
;
BLST:   MOV AL, [BX].RHUCD;             Set pointer in the RH to
        XOR AH, AH;                      the BPB pointer table. 
        SHL AX, 1;
        MOV DI, OFFSET BPBPT;
        ADD DI, AX;
        MOV AX, CS:[DI];
        MOV [BX].BRPTRL, AX;            
        MOV [BX].BRPTRH, CS;            
;
        MOV [BX].RHSTAT, 0100H;         Set success status in the
;                                        Request Header.
;
        RET;                            Return.
;
BLDBPB ENDP
; 
; ----- PERFORM READ/WRITE/WRITE-VERIFY REQUEST.
;
;       ENTRY: DS:BX points to the request header.
;
;       EXIT:  The specified operation has been performed.
;              The status word in the RH has been set to the
;               appropriate value.
;              The registers, AX, CX, DX have been destroyed.
;
SREAD   PROC NEAR
        MOV AX, 0;                      Set control code for READ,
        JMP SHORT SIO;
;
SWRITE  PROC NEAR
        MOV AX, 1;                       WRITE,
        JMP SHORT SIO;
;
SVWRT   PROC NEAR
        MOV AX, 0101H;                   WRITE-VERIFY.
;
SIO:    MOV CS:IOCMMD, AX;              Save the control code.
;
SSVBP:  SUB SP, HD_SLTH;                Create room on the stack
        MOV BP, SP;                      for driver parameters.
;
SCLP:   XOR AH, AH;                     Set command code.
        MOV [BP].HCMMD, AX;
;
        MOV [BP].HCTYPE, 1;             Set caller type,
        MOV AL, [BX].RHUCD;              unit number.
        XOR AH, AH;
        MOV [BP].HCNUM, AX;
        MOV AH, AL;
        MOV CS:CNUM, AX;
        MOV AX, [BX].IOSECT;             starting sector number
        MOV [BP].HCSCTL, AX;
        MOV [BP].HCSCTH, 0;
        MOV AX, [BX].IODTAL;             starting data transfer
        MOV [BP].HDTAL, AX;              address.
        MOV AX, [BX].IODTAH;
        MOV [BP].HDTAH, AX;
;
        MOV DX, [BX].IOCNT;             Load and set the number of
        MOV [BP].HCCNT, DX;              sectors to process.
;
SIOLP:  MOV CX, RT_CNT;                 Perform the operation.
SIORT:  CALL HDDRV;                      If an error, occurs
        CMP AX, 0;                       retry several times.
        JE SNXTC;                        If still no success then
        CMP AX, COR_ERR;                 abort.
        JE SNXTC;
        CALL RSCTL;
        LOOP SIORT;
;
        MOV CX, 0;                      Before aborting clear the
        MOV [BX].IOCNT, CX;              sector count and set the
        CALL IDCD;                       status code.
        JMP SIORTN;
;
SNXTC:  MOV AX, CS:IOCMMD;              Determine if the request
        TEST AX, 0100H;                  command was write-verify.
        MOV AX, 0100H;                   If so, then set verify
        JZ SIORTN;                       command and proceed to
        MOV AX, 0004H;                   loop back to verify that
        MOV CS:IOCMMD, AX;               the data was written
        JMP SCLP;                        correctly.
;
SIORTN: MOV [BX].RHSTAT, AX;            Set result status.
;
        ADD SP, HD_SLTH;                Remove the disk drive para-
;                                        meters from the stack.
;
        RET;                            Return.
;
SVWRT   ENDP
SWRITE  ENDP
SREAD   ENDP
;
; ----- HANDLE INVALID COMMAND FROM DOS.
;
NVCMD   PROC NEAR
;
        MOV [BX].RHSTAT, 8003H;         Set Unknown Command status
;                                        in the request header.
;
        RET;                            Return.
;
NVCMD   ENDP
;
; ----- PERFORM RESET TO SASI CONTROLLERS.
;
;       This subroutine will issue a reset to the SASI bus.  Since
;       all of the controllers on the SASI bus are attached to the
;       same reset line all of them will be reset.  When the SASI
;       controllers are reset they will return to the default
;       drive characteristics.
;
;       This subroutine will, after issueing the reset, reissue the
;       drive characteristics to all of the SASI controllers in-
;       stalled.
;
;       ENTRY:  No specific entry conditions.
;
;       EXIT:   The SASI controllers have been issued a reset
;                command and an attempt has been made to reissue
;                the drive characteristics to each controller.
;
RSCTL   PROC NEAR
;
        PUSH AX;                        Save work registers.
;
        CALL IRSASI;                    Issue the reset command.
;
        MOV AX, CS:CNUM;                Save the current controller
        PUSH AX;                         number.
;
        MOV AL, CS:DUCNT;               Load the number of con-
;                                        trollers configured.
;
        XOR AH, AH;                     Set controller zero as the
;                                        starting controller to
;                                        reissue the drive char-
;                                        acteristics.
;
RSLP:   MOV CS:CNUM, AX;                Reissue the drive char-
        CALL ISDRVCHAR;                  acteristics to controller.
;
        MOV AX, CS:CNUM;                Loop for all controllers.
        INC AH;
        DEC AL;
        JNZ RSLP;
;
        POP AX;                         Restore the controller number
        MOV CS:CNUM, AX;                 set on entry.
;
        POP AX;                         Restore work register and
        RET;                             return.
;
RSCTL   ENDP
;
;
; ----- TRANSLATE ERROR CODE FROM DRIVER TO DOS FORMAT.
;
;       ENTRY: AX contains the error code from HDDRV.
;
;       EXIT:  AX contains the appropriate DOS error code to be
;               place in the Request Header.
;
;
;          HDDRV  DOS                 Description (HDDRV).
ECTAB   DB  02H,  06H;          No seek complete from disk drive.
        DB  03H,  0AH;          Write fault from disk drive.
        DB  04H,  02H;          Drive not ready after select.
        DB  06H,  06H;          Track 0 not found.
        DB  07H,  00H;          Data protected.
        DB  10H,  0BH;          ID field read error.
        DB  11H,  04H;          Uncorrectable data error.
        DB  12H,  0BH;          Address mark not found.
        DB  14H,  08H;          Target sector not found.
        DB  15H,  06H;          Seek error.
        DB  19H,  0BH;          Bad track flag detected.
        DB  32H,  04H;          ECC diagnostic failure.
        DB  50H,  07H;          Media change.
        DB  88H,  0AH;          Verify error.
        DB  -1;                 Table terminator.
;
IDCD    PROC NEAR;
;
        PUSH BX;                        Save work register.
;
        MOV BX, OFFSET ECTAB;           Set pointer to translation
;                                        table.
;
ICMPLP: CMP AL, CS:[BX];                Scan the translation table
        JNE IDCDN;                       for error code from HDDRV.
        CMP BYTE PTR CS:[BX], -1;
        JE IDCDDF;
        INC BX;
        INC BX;
        JMP ICMPLP;
;
IDCDN:  MOV AL, CS:[BX]+1;              If error code was found then
        MOV AH, 80H;                     load the DOS error code,
        JMP IDCRTN;                      set error bit, and exit.
;
IDCDDF: MOV AX, 800CH;                  If the error code was not
;                                        found then set the DOS
;                                        General Failure error code.
;
IDCRTN: POP BX;                         Restore work register and
        RET;                             return.
;
IDCD    ENDP
;
SASI    ENDP
;
PROG    ENDS
;
; ----- REVISION HISTORY. -----------------------------------------*
;
;       ORIGINAL: November 24, 1985  Mike Dougherty
;
;       REV 1.1:  January 15, 1985  Mike Dougherty
;       Added code to the Media Test function that resets the SASI
;       controller on an error and performs retries.  Also checked
;       for Correctable Data Error and did not report such errors to
;       DOS.
;
        END
